/*  $Id: tradext.c,v 1.33 2001/05/25 16:54:17 kondou Exp $
**
**  Storage manager module for traditional spool format.
*/

#include "config.h"
#include "clibrary.h"
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <syslog.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/uio.h>

#ifndef MAP_FAILED
# define MAP_FAILED     (caddr_t) -1
#endif

/* Needed for htonl() and friends on AIX 4.1. */
#include <netinet/in.h>

#include "inn/qio.h"
#include "libinn.h"
#include "macros.h"
#include "paths.h"

#include "methods.h"
#include "../tradspool/trad.h"
#include "tradext.h"

typedef int TEEXT;
#define TE_CROSS   01   /* if an article is crossposted */
#define TE_EXPIRE  02   /* if an article has an Expires: header */
#define TE_MAXVAL  03   /* max value of the extension */
#define TE_BITS     2   /* number of bits required to store extension */
#define TE_EXTLEN   2   /* Room for "." and the octal value of TE_MAXVAL */

TRADDATA *TEdata;

/* Init routine, called by SMinit */

bool
tradext_init(SMATTRIBUTE *attr) {
    if (attr == NULL) {
	syslog(L_ERROR, "tradext: attr is NULL");
	SMseterror(SMERR_INTERNAL, "attr is NULL");
	return FALSE;
    }
    if (!innconf->storeonxref) {
	syslog(L_ERROR, "tradext: storeonxref needs to be true");
	SMseterror(SMERR_INTERNAL, "storeonxref needs to be true");
	return FALSE;
    }
    attr->selfexpire = FALSE;
    attr->expensivestat = TRUE;

    sprintf(TEdata->method, "tradext");
    sprintf(TEdata->dbname, "tradext.map");
    sprintf(TEdata->newdbname, "tradext.map.new");
    TEdata->dbreload = 600;
    TEdata->updated = FALSE;
    TEdata->maxng = 0;

    return TCInitNGTable(TEdata);
}

/* Make a token for an article given the primary newsgroup name and article # */
static TOKEN
MakeToken(char *ng, unsigned long artnum, TEEXT artext, STORAGECLASS class) {
    TOKEN token;
    NGTENT *ngtp;
    unsigned long ngnum, num, maxart;

    memset(&token, '\0', sizeof(token));

    token.type = TOKEN_TRADEXT;
    token.class = class;

    /* 
    ** if not already in the NG Table, be sure to add this ng! This way we
    ** catch things like newsgroups added since startup. 
    */
    if ((ngtp = TCFindNGByName(TEdata, ng)) == NULL) {
	TCAddNG(TEdata, ng, 0);
	TCDumpDB(TEdata); /* flush to disk so other programs can see the change */
	ngtp = TCFindNGByName(TEdata, ng);
    } 

    artnum = htonl(artnum);
    maxart = ~(0)>>TE_BITS;
    if (artnum > maxart) {
        syslog(L_FATAL, "tradext: found %lu articles in a newsgroup, bailing out", artnum);
        SMseterror(SMERR_INTERNAL, "found %lu articles in a newsgroup, bailing out");
        token.type = TOKEN_EMPTY;
        return token;
    }
    num = artnum | artext;
    ngnum = ngtp->ngnumber;
    ngnum = htonl(ngnum);

    memcpy(token.token, &ngnum, sizeof(ngnum));
    memcpy(&token.token[sizeof(ngnum)], &num, sizeof(num));
    return token;
}

/* 
** Convert a token back to a pathname. 
*/
static char *
TokenToPath(TOKEN token) {
    unsigned long ngnum;
    unsigned long artnum, num;
    TEEXT artext;
    char *ng, *path;

    TCCheckNeedReloadDB(TEdata); 

    memcpy(&ngnum, &token.token[0], sizeof(ngnum));
    memcpy(&num, &token.token[sizeof(ngnum)], sizeof(num));
    num = ntohl(num);
    artext = num & TE_MAXVAL;
    artnum = num>>TE_BITS;
    ngnum = ntohl(ngnum);

    ng = TCFindNGByNum(TEdata, ngnum);
    if (ng == NULL) return NULL;

    path = NEW(char, strlen(ng)+20+strlen(innconf->patharticles)+TE_EXTLEN);
    sprintf(path, "%s/%s/%lu.%o", innconf->patharticles, ng, artnum, artext);
    return path;
}

TOKEN
tradext_store(const ARTHANDLE article, const STORAGECLASS class) {
    char **xrefs;
    char *xrefhdr;
    TOKEN token;
    unsigned int numxrefs;
    char *ng, *p, *onebuffer;
    unsigned long artnum;
    TEEXT artext;
    char *path, *linkpath, *dirname;
    int fd;
    int i;
    char *nonwfarticle; /* copy of article converted to non-wire format */
    int nonwflen, used;

    xrefhdr = article.groups;
    if ((xrefs = CrackXref(xrefhdr, &numxrefs)) == NULL || numxrefs == 0) {
	token.type = TOKEN_EMPTY;
	SMseterror(SMERR_UNDEFINED, "bogus Xref: header");
	if (xrefs != NULL)
	    DISPOSE(xrefs);
	return token;
    }

    if ((p = strchr(xrefs[0], ':')) == NULL) {
	token.type = TOKEN_EMPTY;
	SMseterror(SMERR_UNDEFINED, "bogus Xref: header");
	for (i = 0 ; i < numxrefs; ++i) DISPOSE(xrefs[i]);
	DISPOSE(xrefs);
	return token;
    }
    *p++ = '\0';
    ng = xrefs[0];
    DeDotify(ng);
    artnum = atol(p);
    artext = 0;
    if (numxrefs > 1)
        artext &= TE_CROSS;
    if (article.expires > 0)
        artext &= TE_EXPIRE;

    token = MakeToken(ng, artnum, artext, class);

    path = NEW(char, strlen(innconf->patharticles) + strlen(ng) + 32);
    sprintf(path, "%s/%s/%lu.%o", innconf->patharticles, ng, artnum, artext);

    /* following chunk of code boldly stolen from timehash.c  :-) */
    if ((fd = open(path, O_CREAT|O_EXCL|O_WRONLY, ARTFILE_MODE)) < 0) {
	p = strrchr(path, '/');
	*p = '\0';
	if (!MakeDirectory(path, TRUE)) {
	    syslog(L_ERROR, "tradext: could not make directory %s %m", path);
	    token.type = TOKEN_EMPTY;
	    DISPOSE(path);
	    SMseterror(SMERR_UNDEFINED, NULL);
	    for (i = 0 ; i < numxrefs; ++i) DISPOSE(xrefs[i]);
	    DISPOSE(xrefs);
	    return token;
	} else {
	    *p = '/';
	    if ((fd = open(path, O_CREAT|O_EXCL|O_WRONLY, ARTFILE_MODE)) < 0) {
		SMseterror(SMERR_UNDEFINED, NULL);
		syslog(L_ERROR, "tradext: could not open %s %m", path);
		token.type = TOKEN_EMPTY;
		DISPOSE(path);
		for (i = 0 ; i < numxrefs; ++i) DISPOSE(xrefs[i]);
		DISPOSE(xrefs);
		return token;
	    }
	}
    }
    if (innconf->wireformat) {
	if ((xwritev(fd, article.iov, article.iovcnt)) != article.len) {
	    SMseterror(SMERR_UNDEFINED, NULL);
	    syslog(L_ERROR, "tradext error writing %s %m", path);
	    close(fd);
	    token.type = TOKEN_EMPTY;
	    unlink(path);
	    DISPOSE(path);
	    for (i = 0 ; i < numxrefs; ++i) DISPOSE(xrefs[i]);
	    DISPOSE(xrefs);
	    return token;
	}
    } else {
	onebuffer = (char *)xmalloc(article.len);
	for (used = i = 0 ; i < article.iovcnt ; i++) {
	    memcpy(&onebuffer[used], article.iov[i].iov_base, article.iov[i].iov_len);
	    used += article.iov[i].iov_len;
	}
	nonwfarticle = FromWireFmt(onebuffer, used, &nonwflen);
	free(onebuffer);
	if (write(fd, nonwfarticle, nonwflen) != nonwflen) {
	    DISPOSE(nonwfarticle);
	    SMseterror(SMERR_UNDEFINED, NULL);
	    syslog(L_ERROR, "tradext error writing %s %m", path);
	    close(fd);
	    token.type = TOKEN_EMPTY;
	    unlink(path);
	    DISPOSE(path);
	    for (i = 0 ; i < numxrefs; ++i) DISPOSE(xrefs[i]);
	    DISPOSE(xrefs);
	    return token;
	}
	DISPOSE(nonwfarticle);
    }
    close(fd);

    /* 
    ** blah, this is ugly.  Have to make symlinks under other pathnames for
    ** backwards compatiblility purposes. 
    */

    if (numxrefs > 1) {
	for (i = 1; i < numxrefs ; ++i) {
	    if ((p = strchr(xrefs[i], ':')) == NULL) continue;
	    *p++ = '\0';
	    ng = xrefs[i];
	    DeDotify(ng);
	    artnum = atol(p);

	    linkpath = NEW(char, strlen(innconf->patharticles) + strlen(ng) + 32);
	    sprintf(linkpath, "%s/%s/%lu.%o", innconf->patharticles, ng, artnum, artext);
	    if (link(path, linkpath) < 0) {
		p = strrchr(linkpath, '/');
		*p = '\0';
		dirname = COPY(linkpath);
		*p = '/';
		if (!MakeDirectory(dirname, TRUE) || link(path, linkpath) < 0) {
#if !defined(HAVE_SYMLINK)
		    SMseterror(SMERR_UNDEFINED, NULL);
		    syslog(L_ERROR, "tradext: could not link %s to %s %m", path, linkpath);
		    token.type = TOKEN_EMPTY;
		    DISPOSE(dirname);
		    DISPOSE(linkpath);
		    DISPOSE(path);
		    for (i = 0 ; i < numxrefs; ++i) DISPOSE(xrefs[i]);
		    DISPOSE(xrefs);
		    return token;
#else
		    if (symlink(path, linkpath) < 0) {
			SMseterror(SMERR_UNDEFINED, NULL);
			syslog(L_ERROR, "tradext: could not symlink %s to %s %m", path, linkpath);
			token.type = TOKEN_EMPTY;
			DISPOSE(dirname);
			DISPOSE(linkpath);
			DISPOSE(path);
			for (i = 0 ; i < numxrefs; ++i) DISPOSE(xrefs[i]);
			DISPOSE(xrefs);
			return token;
		    }
#endif  /* !defined(HAVE_SYMLINK) */
		}
		DISPOSE(dirname);
	    }
	    DISPOSE(linkpath);
	}
    }
    DISPOSE(path);
    for (i = 0 ; i < numxrefs; ++i) DISPOSE(xrefs[i]);
    DISPOSE(xrefs);
    return token;
}

ARTHANDLE *
tradext_retrieve(const TOKEN token, const RETRTYPE amount) {
    char *path;
    ARTHANDLE *art;
    static TOKEN ret_token;

    if (token.type != TOKEN_TRADEXT) {
	SMseterror(SMERR_INTERNAL, NULL);
	return NULL;
    }

    if ((path = TokenToPath(token)) == NULL) {
	SMseterror(SMERR_NOENT, NULL);
	return NULL;
    }
    if ((art = TCOpenArticle(TEdata, path, amount)) != (ARTHANDLE *)NULL) {
        ret_token = token;
        art->token = &ret_token;
    }
    DISPOSE(path);
    return art;
}

void
tradext_freearticle(ARTHANDLE *article) {
    PRIV_TRAD *private;

    if (!article) return;

    if (article->private) {
	private = (PRIV_TRAD *) article->private;
	if (private->mmapped) {
#if defined(MADV_DONTNEED) && defined(HAVE_MADVISE)
	    madvise(private->artbase, private->artlen, MADV_DONTNEED);
#endif
	    munmap(private->artbase, private->artlen);
	} else {
	    DISPOSE(private->artbase);
	}
	if (private->curdir) {
	    closedir(private->curdir);
	}
	DISPOSE(private->curdirname);
    }
}

bool 
tradext_cancel(TOKEN token) {
    char **xrefs;
    char *xrefhdr;
    ARTHANDLE *article;
    unsigned int numxrefs;
    char *ng, *p;
    char *path, *linkpath;
    int i;
    bool result = TRUE;
    unsigned long artnum;
    TEEXT artext;

    /*
    ** I have a feeling SMERR_UNDEFINED below should actually be SMERR_NOENT.
    ** At least in some cases. Leaving it be for now; Peppe
    */

    if ((path = TokenToPath(token)) == NULL) {
	SMseterror(SMERR_UNDEFINED, NULL);
	DISPOSE(path);
	return FALSE;
    }

    p = strrchr(path, '/');
    p = strrchr(p, '.');
    if (strlen(p) > TE_EXTLEN) {
	SMseterror(SMERR_UNDEFINED, NULL);
	DISPOSE(path);
        DISPOSE(p);
	return FALSE;
    }
    artext = atol(++p);
    DISPOSE(p);

    /*
    **  Ooooh, this is gross.  To find the symlinks pointing to this article,
    ** we open the article and grab its Xref line (since the token isn't long
    ** enough to store this info on its own).   This is *not* going to do 
    ** good things for performance of fastrm...  -- rmtodd
    **
    ** With tradext we only open it if it is crossposted... -- Peppe
    */
    if (artext & TE_CROSS) {
        if ((article = TCOpenArticle(TEdata, path, RETR_HEAD)) == NULL) {
            DISPOSE(path);
            SMseterror(SMERR_UNDEFINED, NULL);
            return FALSE;
        }

        xrefhdr = (char *)HeaderFindMem(article->data, article->len, "Xref", 4);
        if (xrefhdr == NULL) {
            /* for backwards compatibility; there is no Xref unless crossposted
               for 1.4 and 1.5 */
            if (unlink(path) < 0) result = FALSE;
            DISPOSE(path);
            tradext_freearticle(article);
            return result;
        }

        if ((xrefs = CrackXref(xrefhdr, &numxrefs)) == NULL || numxrefs == 0) {
            DISPOSE(path);
            tradext_freearticle(article);
            SMseterror(SMERR_UNDEFINED, NULL);
            return FALSE;
        }

        tradext_freearticle(article);
        for (i = 1 ; i < numxrefs ; ++i) {
            if ((p = strchr(xrefs[i], ':')) == NULL) continue;
            *p++ = '\0';
            ng = xrefs[i];
            DeDotify(ng);
            artnum = atol(p);

            linkpath = NEW(char, strlen(innconf->patharticles) + strlen(ng) + 32);
            sprintf(linkpath, "%s/%s/%lu.%o, innconf->patharticles, ng, artnum, artext);
            /* hmm, do we want to abort this if one of the symlink unlinks fails? */
            if (unlink(linkpath) < 0) result = FALSE;
            DISPOSE(linkpath);
        }
    }
    if (unlink(path) < 0) result = FALSE;
    DISPOSE(path);
    for (i = 0 ; i < numxrefs ; ++i) DISPOSE(xrefs[i]);
    DISPOSE(xrefs);
    return result;
}

   
/*
** Find entries for possible articles in dir. "dir" (directory name "dirname").
** The dirname is needed so we can do stats in the directory to disambiguate
** files from symlinks and directories.
*/

static struct dirent *
FindDir(DIR *dir, char *dirname) {
    struct dirent *de;
    int i;
    bool flag;
    char *path;
    struct stat sb;
    unsigned char namelen;

    while ((de = readdir(dir)) != NULL) {
	namelen = strlen(de->d_name);
	for (i = 0, flag = TRUE ; i < namelen ; ++i) {
	    if (!(isdigit(de->d_name[i]) || de->d_name[i] == '.')) {
		flag = FALSE;
		break;
	    }
	}
	if (!flag) continue; /* if not all digits or '.', skip this entry. */

	path = NEW(char, strlen(dirname)+namelen+2);
	strcpy(path, dirname);
	strcat(path, "/");
	strncpy(&path[strlen(dirname)+1], de->d_name, namelen);
	path[strlen(dirname)+namelen+1] = '\0';

	if (lstat(path, &sb) < 0) {
	    DISPOSE(path);
	    continue;
	}
	DISPOSE(path);
	if (!S_ISREG(sb.st_mode)) continue;
	return de;
    }
    return NULL;
}

ARTHANDLE *tradext_next(const ARTHANDLE *article, const RETRTYPE amount) {
    PRIV_TRAD priv;
    PRIV_TRAD *newpriv;
    char *path, *linkpath;
    struct dirent *de;
    ARTHANDLE *art;
    unsigned long artnum;
    TEEXT artext;
    int i;
    static TOKEN token;
    unsigned char namelen;
    char **xrefs;
    char *xrefhdr, *ng, *p, *expires, *x, *fname;
    unsigned int numxrefs;
    STORAGE_SUB	*sub;

    if (article == NULL) {
	priv.ngtp = NULL;
	priv.curdir = NULL;
	priv.curdirname = NULL;
	priv.nextindex = -1;
    } else {
	priv = *(PRIV_TRAD *) article->private;
	DISPOSE(article->private);
	DISPOSE((void*)article);
	if (priv.artbase != NULL) {
	    if (priv.mmapped) {
#if defined(MADV_DONTNEED) && defined(HAVE_MADVISE)
		madvise(priv.artbase, priv.artlen, MADV_DONTNEED);
#endif
		munmap(priv.artbase, priv.artlen);
	    } else {
		DISPOSE(priv.artbase);
	    }
	}
    }

    while (!priv.curdir || ((de = FindDir(priv.curdir, priv.curdirname)) == NULL)) {
	if (priv.curdir) {
	    closedir(priv.curdir);
	    priv.curdir = NULL;
	    DISPOSE(priv.curdirname);
	    priv.curdirname = NULL;
	}

	/*
	** advance ngtp to the next entry, if it exists, otherwise start 
	** searching down another ngtable hashchain. 
	*/
	while (priv.ngtp == NULL || (priv.ngtp = priv.ngtp->next) == NULL) {
	    /*
	    ** note that at the start of a search nextindex is -1, so the inc.
	    ** makes nextindex 0, as it should be.
	    */
	    priv.nextindex++;
	    if (priv.nextindex >= NGT_SIZE) {
		/* ran off the end of the table, so return. */
		return NULL;
	    }
	    priv.ngtp = TEdata->table[priv.nextindex];
	    if (priv.ngtp != NULL)
		break;
	}

	priv.curdirname = NEW(char, strlen(innconf->patharticles)+strlen(priv.ngtp->ngname)+2);
	sprintf(priv.curdirname, "%s/%s",innconf->patharticles,priv.ngtp->ngname);
	priv.curdir = opendir(priv.curdirname);
    }

    namelen = strlen(de->d_name);
    path = NEW(char, strlen(priv.curdirname) + 2 + namelen);
    strcpy(path, priv.curdirname);
    strcat(path, "/");
    i = strlen(priv.curdirname);
    strncpy(&path[i+1], de->d_name, namelen);
    path[i+namelen+1] = '\0';
    /* get the article number while we're here, we'll need it later. */
    fname = COPY(&p[i+1]);
    p = strrchr(fname, '.');
    *p++ = '\0';
    artext = atol(p);
    artnum = atol(fname);
    DISPOSE(fname);

    art = TCOpenArticle(TEdata, path, amount);
    if (art == (ARTHANDLE *)NULL) {
	art = NEW(ARTHANDLE, 1);
	art->type = TOKEN_TRADEXT;
	art->data = NULL;
	art->len = 0;
	art->private = (void *)NEW(PRIV_TRAD, 1);
	art->expires = 0;
	art->groups = NULL;
	art->groupslen = 0;
	newpriv = (PRIV_TRAD *) art->private;
	newpriv->artbase = NULL;
    } else {
	/* Skip linked (not symlinked) crossposted articles.

           This algorithm is rather questionable; it only works if the first
           group/number combination listed in the Xref header is the
           canonical path.  This will always be true for spools created by
           this implementation, but for traditional INN 1.x servers,
           articles are expired indepedently from each group and may expire
           out of the first listed newsgroup before other groups.  This
           algorithm will orphan such articles, not adding them to history.

           The bit of skipping articles by setting the length of the article
           to zero is also rather suspect, and I'm not sure what
           implications that might have for the callers of SMnext.

           Basically, this whole area really needs to be rethought. */

        /* We only need to do this for crossposted articles */
        if (artext & TE_CROSS) {
            xrefhdr = (char *)HeaderFindMem(art->data, art->len, "Xref", 4);
            if (xrefhdr != NULL) {
                if ((xrefs = CrackXref(xrefhdr, &numxrefs)) == NULL || numxrefs == 0) {
                    art->len = 0;
                } else {
                    /* assumes first one is the original */
                    if ((p = strchr(xrefs[0], ':')) != NULL) {
                        *p++ = '\0';
                        ng = xrefs[0];
                        DeDotify(ng);
                        artnum = atol(p);

                        linkpath = NEW(char, strlen(innconf->patharticles) + strlen(ng) + 32);
                        sprintf(linkpath, "%s/%s/%lu.%o", innconf->patharticles, ng, artnum, artext);
                        if (strcmp(path, linkpath) != 0) {
                            /* this is linked article, skip it */
                            art->len = 0;
                        }
                        DISPOSE(linkpath);
                    }
                }
                for (i = 0 ; i < numxrefs ; ++i) DISPOSE(xrefs[i]);
                DISPOSE(xrefs);
                if (innconf->storeonxref) {
                    /* skip path element */
                    if ((xrefhdr = strchr(xrefhdr, ' ')) == NULL) {
                        art->groups = NULL;
                        art->groupslen = 0;
                    } else {
                        for (xrefhdr++; *xrefhdr == ' '; xrefhdr++);
                        art->groups = xrefhdr;
                        for (p = xrefhdr ; (*p != '\n') && (*p != '\r') ; p++);
                        art->groupslen = p - xrefhdr;
                    }
                }
            } else if (innconf->storeonxref) {
                art->groups = NULL;
                art->groupslen = 0;
            }
        } else {
            art->groups = ng;
            for (p = ng ; (*p != '\n') && (*p != '\r') ; p++);
            art->groupslen = p - ng;
        }
	if (!innconf->storeonxref) {
	    if ((ng = (char *)HeaderFindMem(art->data, art->len, "Newsgroups", 10)) == NULL) {
		art->groups = NULL;
		art->groupslen = 0;
	    } else {
		art->groups = ng;
		for (p = ng ; (*p != '\n') && (*p != '\r') ; p++);
		art->groupslen = p - ng;
	    }
	}
        if (artext & TE_EXPIRE) {
            if ((expires = (char *)HeaderFindMem(art->data, art->len, "Expires", 7)) == NULL) {
                art->expires = 0;
            } else {
                /* optionally parse expire header */
                for (p = expires + 1; (*p != '\n') && (*(p - 1) != '\r'); p++);
                x = NEW(char, p - expires);
                memcpy(x, expires, p - expires - 1);
                x[p - expires - 1] = '\0';

                art->expires = parsedate(x, NULL);
                if (art->expires == -1)
                    art->expires = 0;
                else
                    art->expires -= time(0);
                DISPOSE(x);
            }
        } else {
            art->expires = 0;
        }
	/* for backwards compatibility; assumes no Xref unless crossposted
	   for 1.4 and 1.5: just fall through */
    }
    newpriv = (PRIV_TRAD *) art->private;
    newpriv->nextindex = priv.nextindex;
    newpriv->curdir = priv.curdir;
    newpriv->curdirname = priv.curdirname;
    newpriv->ngtp = priv.ngtp;
    
    if ((sub = SMgetsub(*art)) == NULL || sub->type != TOKEN_TRADEXT) {
	/* maybe storage.conf is modified, after receiving article */
	token = MakeToken(priv.ngtp->ngname, artnum, artext, 0);

        /* Only log an error if art->len is non-zero, since otherwise we get
           all the ones skipped via the hard-link skipping algorithm
           commented above. */
        if (art->len > 0)
            syslog(L_ERROR, "tradext: can't determine class: %s: %s",
                   TokenToText(token), SMerrorstr);
    } else {
	token = MakeToken(priv.ngtp->ngname, artnum, artext, sub->class);
    }
    art->token = &token;
    DISPOSE(path);
    return art;
}

bool tradext_ctl(PROBETYPE type, TOKEN *token, void *value) {
    struct artngnum *ann;
    unsigned long ngnum;
    unsigned long artnum, artext;
    char *ng;

    switch (type) { 
    case SMARTNGNUM:
	if ((ann = (struct artngnum *)value) == NULL)
	    return FALSE;
	TCCheckNeedReloadDB(TEdata);
	memcpy(&ngnum, &token->token[0], sizeof(ngnum));
	memcpy(&artnum, &token->token[sizeof(ngnum)], sizeof(artext));
        /* remember to shift off the extension from the article number */
	artnum = ntohl(artnum) >> TE_BITS;
	ngnum = ntohl(ngnum);
	ng = TCFindNGByNum(TEdata, ngnum);
	if (ng == NULL) return FALSE;
	ann->groupname = COPY(ng);
	ann->artnum = (ARTNUM)artnum;
	return TRUE;
    default:
	return FALSE;
    }       
}

bool tradext_flushcacheddata(FLUSHTYPE type) {
    return TRUE;
}

void tradext_printfiles(FILE *file, TOKEN token, char **xref, int ngroups) {
    unsigned long ngnum, artext;
    TEEXT ext;
    int i;
    char *path, *p;

    memcpy(&artext, &token.token[sizeof(ngnum)], sizeof(artext));
    ext = ntohl(artext) & TE_MAXVAL;
    for (i = 0; i < ngroups; i++) {
        path = xstrdup(xref[i]);
        for (p = path; *p != '\0'; p++)
            if (*p == '.' || *p == ':')
                *p = '/';
        fprintf(file, "%s.%o\n", path, ext);
        free(path);
    }
}

void
tradext_shutdown(void) {
    TCDumpDB(TEdata);
    TCFreeNGTree(TEdata);
}
